#####
###  MailBoxer - a mailinglist/newsletter/mailarchive-framework for ZOPE
##   Copyright (C) 2002 Maik Jablonski (maik.jablonski@uni-bielefeld.de)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
##   along with this program; if not, write to the Free Software
###  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#####

# $Id: MailBoxer.py,v 1.19 2002/12/23 14:30:45 mjablonski Exp $

import StringIO, re, difflib, random, md5
import cgi, smtplib, rfc822, multifile, mimetools, mimify
from DateTime import DateTime
from Globals import InitializeClass, DTMLFile
from AccessControl import ClassSecurityInfo
from OFS.Folder import Folder
from zLOG import LOG, INFO
from MailBoxerTemplates import MailBoxerTemplates

manage_addMailBoxerForm = DTMLFile('dtml/manage_addMailBoxerForm', globals())

def manage_addMailBoxer (self, id,
                title='', 
                mailto='', 
                moderator=[],
                moderated=0,
                archived='not archived',
                mtahosts=[],
                smtphost='127.0.0.1',
                REQUEST=None
            ):
    """ Add a new MailBoxer to current ObjectManager """

    mb = MailBoxer(id, title, mailto, moderator, moderated, archived, mtahosts)
    self._setObject(id,mb)
    mb=self._getOb(id)

    # Add a MailHost and a ZCatalog
    mb.manage_addFolder('archive','Archive')
    mb.manage_addProduct['MailHost'].manage_addMailHost('MailHost', 
                                                         smtp_host=smtphost)
    
    mb.manage_addProduct['ZCatalog'].manage_addZCatalog('Catalog','Catalog')
    mb.Catalog.manage_addIndex('mailDate', 'FieldIndex')
    mb.Catalog.manage_addIndex('mailFrom', 'TextIndex')
    mb.Catalog.manage_addIndex('mailSubject','TextIndex')
    mb.Catalog.manage_addIndex('mailBody', 'TextIndex')
    
    # Add dtml-templates
    for (id, data) in MailBoxerTemplates.items():
        mb.addDTMLMethod(id, file=data)
        
    # Redirect if requested TTW
    if REQUEST:
        return self.manage_main(self,REQUEST,update_menu=1)
        

class MailBoxer(Folder):
    """ a folder with maillist-managing-features """

    security = ClassSecurityInfo()
    
    meta_type = 'Mail Boxer'

    archive_options = ['not archived', 'plain text', 'with attachments']

    # The _hashkey is used for the md5-checksum for verifying (un)subscriptions. 
    # It will be reset every time you restart your Zope-Server.
    _hashkey   = str(random.random())
    
    # init properties
    title = ''
    mailto = ''
    maillist = []
    moderator = []
    moderated = 0
    plainmail = 0
    archived = ''
    subscribe = ''
    unsubscribe = ''
    mtahosts = []
    spamlist = []
    atmask = ''
    sniplist = ''
    catalog = ''
    xmailer = ''
    
    _properties=(
        {'id':'title', 'type':'string', 'mode':'w'},
        {'id':'mailto', 'type':'string', 'mode':'w'},
        {'id':'maillist', 'type':'lines', 'mode':'w'},
        {'id':'moderator', 'type':'lines', 'mode':'w'},
        {'id':'moderated', 'type':'boolean', 'mode':'w'},
        {'id':'plainmail', 'type':'boolean', 'mode':'w'},
        {'id':'archived', 'type':'selection','mode':'w', 
                      'select_variable':'archive_options'},
        {'id':'subscribe', 'type':'string', 'mode':'w'},
        {'id':'unsubscribe','type':'string', 'mode':'w'},
        {'id':'mtahosts', 'type':'tokens', 'mode':'w'},
        {'id':'spamlist', 'type':'lines', 'mode':'w'},
        {'id':'atmask', 'type':'string', 'mode':'w'},
        {'id':'sniplist', 'type':'lines', 'mode':'w'},
        {'id':'catalog', 'type':'string', 'mode':'w'},
        {'id':'xmailer', 'type':'string', 'mode':'w'}
    )

    def __init__(self, id, title, mailto, moderator, moderated, archived, mtahosts):
        """ setup a MailBoxer with reasonable defaults """
        
        self.id = id
        self.title = title
        self.mailto = mailto
        self.maillist = []
        self.moderator = moderator
        self.moderated = moderated
        self.plainmail = 0
        self.archived = archived
        self.subscribe = 'subscribe'
        self.unsubscribe = 'unsubscribe'
        self.mtahosts = mtahosts
        self.spamlist = []
        self.atmask = '(at)'
        self.sniplist = [r'(\n>.*)+|(\n.>.*)+', 
                         r'(?s)\n-- .*', 
                         r'(?s)\n_+\n.*']
        self.catalog = 'Catalog'
        self.xmailer = 'MailBoxer'

    
    security.declarePrivate('_unpackMultifile')
    def _unpackMultifile(self,obj,multifile):
        """ unpack body and attachments; (TextBody, ContentType, HtmlBody) """

        TextBody = HtmlBody = ContentType = ''
        
        msg = mimetools.Message(multifile)
        maintype = msg.getmaintype()
        subtype = msg.getsubtype()
        name = msg.getparam('name')

        # Recurse over all nested multiparts
        if maintype == 'multipart':
            multifile.push(msg.getparam('boundary'))
            multifile.readlines()
            while not multifile.last:
                multifile.next()
                
                (tmpTextBody, tmpContentType, tmpHtmlBody) = \
                                           self._unpackMultifile(obj, multifile)
                
                # Return ContentType only for the plain-body of a mail
                if tmpContentType:
                    TextBody = tmpTextBody
                    ContentType = tmpContentType
                
                if tmpHtmlBody:
                    HtmlBody = tmpHtmlBody
                    
            multifile.pop()
            return (TextBody,ContentType,HtmlBody)

        # Process MIME-encoded data
        plainfile = StringIO.StringIO()

        try:
            mimetools.decode(multifile,plainfile,msg.getencoding())
        # unknown or no encoding? 7bit, 8bit or whatever... copy literal
        except ValueError:
            mimetools.copyliteral(multifile,plainfile)

        body = plainfile.getvalue()
        plainfile.close()
    
        # Get plain text
        if maintype == 'text' and subtype == 'plain' and not name:
            TextBody = body
            ContentType = msg.get('content-type', 'text/plain')
            
        # Decode the rest
        else:
            # No name? This should be the html-body...
            if not name:
                name = '%s.%s' % (maintype,subtype)
                HtmlBody = body

            if obj and self.archived=='with attachments':
           
                id = DateTime().millis()
                # to be sure: test and search for a free id...
                while hasattr(obj,str("%s.%s" % (str(id), subtype))):  
                    id = id + 1
                    
                id = "%s.%s" % (str(id), subtype)
                obj.manage_addFile(id, title = name, file = body, 
                                   content_type = maintype + '/' + subtype)

        return (TextBody, ContentType, HtmlBody)


    security.declareProtected('Add Folders','manage_addMail')
    def manage_addMail(self,Mail):
        """ store mail & attachments in a folder and return it """

        (mailHeader, mailBody) = self.splitMail(Mail)

        # taste the mail-date => take our own if it's bad
        try:
            time = DateTime(mailHeader['date'])
            if time >= DateTime():
                time = DateTime()
        except: 
            time = DateTime()
        
        # now let's create the date-path (yyyy/yyyy-mm) 
        year  = str(time.year())                  # yyyy
        month = "%s-%s" % (year, str(time.mm()))  # yyyy-mm

        archive = getattr(self,'archive')
        
        # do we have a year folder already?
        if not hasattr(archive, year):
            archive.manage_addFolder(year, title=year)
        yearFolder=getattr(archive, year)

        # do we have a month folder already?
        if not hasattr(yearFolder, month):
            yearFolder.manage_addFolder(month, title=month)
        monthFolder=getattr(yearFolder, month)

        # let's create the mailObject
        mailFolder = monthFolder
        
        Subject = self.mime_decode_header(mailHeader.get('subject',
                                                         'No Subject'))
        From = self.mime_decode_header(mailHeader.get('from','No Sender'))
        Title = "%s / %s" % (Subject, From)

        # maybe it's a reply ?
        if ':' in Subject:
            for currentFolder in monthFolder.objectValues(['Folder']):
                if difflib.get_close_matches(Subject,
                                             [currentFolder.mailSubject]):
                    mailFolder=currentFolder

        # search a free id for the mailobject
        id = time.millis()
        while hasattr(mailFolder,str(id)):  
             id = id + 1
             
        id = str(id)

        mailFolder.manage_addFolder(id, title=Title)
        mailObject = getattr(mailFolder, id)

        # unpack attachments
        (TextBody, ContentType, HtmlBody) =  self._unpackMultifile(mailObject, 
                                                     multifile.MultiFile(
                                                      StringIO.StringIO(Mail)))

        # ContentType is only set for the TextBody
        if ContentType:
            mailBody = TextBody
        else:
            mailBody = self.HtmlToText(HtmlBody)
        
        # and now add some properties to our new mailobject 
        mailObject.manage_addProperty('mailFrom', From, 'string')
        mailObject.manage_addProperty('mailSubject', Subject, 'string')
        mailObject.manage_addProperty('mailDate', time, 'date')
        mailObject.manage_addProperty('mailBody', mailBody, 'text')
        
        # Index the new created mailFolder in the catalog
        if hasattr(self, self.catalog):
            getattr(self, self.catalog).catalog_object(mailObject)
    
        return mailObject

    
    security.declarePublic('splitMail')
    def splitMail(self,Mail):
        """ return (header,body) of a mail given as string """

        msg = mimetools.Message(StringIO.StringIO(Mail))
        
        # Get headers
        mailHeader = {}
        for (key,value) in msg.items():
            mailHeader[key] = value
        
        # Get body
        msg.rewindbody()
        mailBody = msg.fp.read()

        return (mailHeader, mailBody)


    security.declarePublic('securityCheck')
    def securityCheck(self,REQUEST):
        """ check for ip, x-mailer and spam """

        if self.mtahosts:
            if 'HTTP_X_FORWARDED_FOR' in self.REQUEST.environ.keys():
                REMOTE_IP = self.REQUEST.environ['HTTP_X_FORWARDED_FOR']
            else:
                REMOTE_IP = self.REQUEST.environ['REMOTE_ADDR']

            if REMOTE_IP not in self.mtahosts:          
                message = 'Host %s is not allowed' % (REMOTE_IP)
                LOG('MailBoxer', INFO,  message)
                raise(message)

        # mail already processed? this avoids endless loops...
        (mailHeader,mailBody)=self.splitMail(REQUEST['Mail'])

        if mailHeader.get('x-mailer')==self.xmailer:
            message = 'Mail already processed'
            LOG('MailBoxer', INFO, message)
            raise(message)

        # mail is spam? log this and raise an error.
        for regexp in self.spamlist:
            if regexp and re.search(regexp,REQUEST['Mail']):
                message='Spam detected: %s\n\n%s\n' % (regexp,REQUEST['Mail'])
                LOG('MailBoxer', INFO, message)
                raise(message)
                

    security.declarePublic('manage_inboxer')
    def manage_inboxer(self, REQUEST):
        """ wrapper to post directly into archives """

        self.securityCheck(REQUEST)
        self.manage_addMail(REQUEST['Mail'])


    security.declarePublic('manage_listboxer')
    def manage_listboxer(self, REQUEST):
        """ bulks a mail to all members of the list """ 

        self.securityCheck(REQUEST)

        Mail = REQUEST['Mail']
        context = self
        
        # store mail in the archive?
        if self.archived <> 'not archived':
            context = self.manage_addMail(Mail)

        (mailHeader, mailBody) = self.splitMail(Mail)

        # plainmail-mode? get the plain-text of mail...
        if self.plainmail:          
            (TextBody, ContentType, HtmlBody) = self._unpackMultifile(None, 
                                                   multifile.MultiFile(
                                                    StringIO.StringIO(Mail)))
            if ContentType:
                mailHeader['content-type'] = ContentType
                mailBody = TextBody
            else:
                mailHeader['content-type'] = 'text/plain; charset=ISO-8859-1;'
                mailBody = self.HtmlToText(HtmlBody)
        
        # get custom header & footer
        (customHeader, customBody) = self.splitMail(self.mail_header(context,
                                                    REQUEST, title=self.title, 
                                                    mail=mailHeader, 
                                                    body=mailBody).strip())
                                                    
        customFooter = self.mail_footer(context, REQUEST, title=self.title,
                                                          mail=mailHeader, 
                                                          body=mailBody).strip()
            
        
        # Patch msg-headers with customHeaders
        msg = mimetools.Message(StringIO.StringIO(Mail))
        
        for hdr in customHeader.keys():
            if customHeader[hdr]:
                msg[hdr.capitalize()]=customHeader[hdr]
            else:
                del msg[hdr]
        
        newHeader = ''.join(msg.headers)

        # If customBody is not empty, use it as new mailBody
        if customBody.strip():
            mailBody=customBody

        newMail = "%s\r\n%s\r\n%s" % (newHeader, mailBody, customFooter)

        # Get additional Members from mail_out_members-hook
        MailOutMembers = self.maillist[:]
        if hasattr(self,'mail_out_members'):
            MailOutMembers = MailOutMembers + self.mail_out_members()
        
        # bulk mail. return-path is first moderator.
        try:
            smtpserver = smtplib.SMTP(self.MailHost.smtp_host, 
                                       int(self.MailHost.smtp_port))
            smtpserver.sendmail(self.moderator[0], MailOutMembers, newMail)
            smtpserver.quit()
        except:
            LOG('MailBoxer', INFO, 'SMTP-Error')


    security.declarePublic('manage_mailboxer')
    def manage_mailboxer(self, REQUEST):
        """ handles all mails and (un)subscriptions """

        self.securityCheck(REQUEST)
        
        (mailHeader, mailBody) = self.splitMail(REQUEST['Mail'])
        
        # get subject and email-address
        Subject = self.mime_decode_header(mailHeader.get('subject',
                                                         'No Subject'))
        From = self.mime_decode_header(mailHeader.get('from', 'No Sender'))
        (RealName, RealEmail) = rfc822.parseaddr(From)
        
        # lower-case all addresses for comparisons
        LowEmail = RealEmail.lower()

        # Get members from maillist-property
        MailList=[]
        for addr in self.maillist:
            MailList.append(addr.lower())

        # Get additional Members over mail_in_members-hook
        MailInMembers = []
        if hasattr(self, 'mail_in_members'):
            for addr in self.mail_in_members():
                MailInMembers.append(addr.lower())

        # Get moderators
        ModeratorList=[]
        for addr in self.moderator:
            ModeratorList.append(addr.lower())

        # subscription? only subscribe if subscription is enabled.
        if self.subscribe<>'' and re.match('(?i)' + self.subscribe + "|.*: " +
                                                    self.subscribe, Subject):

            if Subject.find(self.pin(From))<>-1:
                if LowEmail not in MailList:
                    self.maillist.append(RealEmail)
                    self.maillist.sort()
                    self.maillist=self.maillist
            self.mail_subscribe(self, REQUEST, mail=mailHeader, body=mailBody)
            return

        # unsubscription? only unsubscribe if unsubscription is enabled...
        if self.unsubscribe<>'' and re.match('(?i)' + self.unsubscribe + 
                                               "|.*: " + self.unsubscribe,
                                               Subject):

            if Subject.find(self.pin(From))<>-1:
                if LowEmail in MailList:
                    index=MailList.index(LowEmail)
                    self.maillist=self.maillist[:index] + self.maillist[index+1:]
            self.mail_unsubscribe(self, REQUEST, mail=mailHeader, body=mailBody)
            return

        # message to a moderated list... relay all mails from a moderator
        if self.moderated:
            if LowEmail in ModeratorList:
                self.mail_handler(self, REQUEST, mail=mailHeader,
                                                        body=mailBody)
            else:
                self.mail_moderator(self, REQUEST, mail=mailHeader,
                                                        body=mailBody)
            return

        # traffic! relay all mails from members and moderators...

        if LowEmail in MailList + MailInMembers + ModeratorList:
            self.mail_handler(self, REQUEST, mail=mailHeader, body=mailBody)
            return

        # if all previous tests fail, it must be an unknown sender.
        self.mail_reply(self, REQUEST, mail=mailHeader, body=mailBody)
        return

    
    security.declarePublic('pin')
    def pin(self, mailFrom):
        """ returns the hex-digits for a randomized md5 """
        
        (RealName, RealEmail) = rfc822.parseaddr(mailFrom)      
        return md5.new(RealEmail.lower() + self._hashkey).hexdigest()
        

    security.declarePublic('mime_decode_header')
    def mime_decode_header(self, strg):
        """ returns the undecoded header; just a public wrapper """
        
        return mimify.mime_decode_header(strg)
        

    security.declarePublic('snipMail')      
    def snipMail(self, strg):
        """ snips all regexps from strg and urlifies it """
        
        for regexp in self.sniplist:
            if regexp: strg = re.sub(regexp,'[...]',strg)

        strg = cgi.escape(strg)
        
        # Urlifing can be done better... but it's quite good
        strg = re.sub(r'(?P<url>http://[-_&;,?:~=%#+/.0-9a-zA-Z]+)', 
                      r'<a href="\g<url>">\g<url></a>', strg)

        return strg.strip()


    security.declarePublic('at')
    def at(self,strg):
        """ replaces all @ with the atmask """
        
        return(strg.replace('@',self.atmask))
        

    security.declarePublic('nextObject')
    def nextObject(self,context):
        """ Returns next Folder in context or None """
           
        parentObjects = context.aq_parent.objectIds('Folder')
        parentObjects.sort()

        if context.getId() in parentObjects:
            parentPosition = parentObjects.index(context.getId())
            
            if parentPosition+1<len(parentObjects):
                return getattr(context,str(parentObjects[parentPosition+1]))
               
            elif context<>self and context.aq_parent<>self:
                if self.nextObject(context.aq_parent):
                    objList = self.nextObject(context.aq_parent).objectValues('Folder')
                    if objList:
                        return objList[0]

        return None


    security.declarePublic('previousObject')
    def previousObject(self,context):
        """ Return previous Folder in context or None """
           
        parentObjects = context.aq_parent.objectIds('Folder')
        parentObjects.sort()
 
        if context.getId() in parentObjects:
            parentPosition = parentObjects.index(context.getId())
                   
            if parentPosition > 0:
                return getattr(context,str(parentObjects[parentPosition-1]))
               
            elif context<>self and context.aq_parent<>self:
                if self.previousObject(context.aq_parent):
                    objList = self.previousObject(context.aq_parent).objectValues('Folder')
                    if objList:
                        return objList[-1]
               
        return None

    security.declarePublic('HtmlToText')
    def HtmlToText(self, html):
        """ convert html to text """
        
        p = HTMLStripper()
        p.feed(html)
        p.close()
        return str(p)
        
# And here we go to the moon... Houston, we have a problem...
InitializeClass(MailBoxer)

# HTMLStripper is inspired by StripTagParser from Dieter Maurer

from sgmllib import SGMLParser
class HTMLStripper(SGMLParser):
    """ remove tags and translate entities """

    from htmlentitydefs import entitydefs

    text = ''

    def handle_data(self, data):
        self.text = self.text + data

    def __str__(self):
        return self.text.strip()
